@@ -4,12 +4,9 @@
 from biobookshelf.main import *
 from biobookshelf import *
 
-pd.options.mode.chained_assignment = None  # default='warn' # to disable worining
 
 from cressp.structural_property_estimation import Estimate_structural_property
-
 from cressp.cross_reactivity_prediction import Predict_T_cell_cross_reactivity, Predict_B_cell_cross_reactivity
-
 from cressp.web_application import Prepare_data_for_web_application
 
 import argparse
@@ -19,13 +16,21 @@ from io import StringIO
 import time
 import math
 
+pd.options.mode.chained_assignment = None  # default='warn' # to disable worining
+import warnings
+warnings.filterwarnings( action = 'ignore' )
+# from scarab.quality_control import Estimate_structural_property
+
+# retrieve a function for logging
+Info = multiprocessing.get_logger( ).info
+
 #     # read dict_blosum62 from the tsv file
 #     df_blosum62 = pd.read_csv( f'{dir_folder_cressp}data/blosum62.tsv.gz', sep = '\t' )
 #     dict_blosum62 = dict( )
 #     for aa_0, aa_1, score in df_blosum62.values : # sould be in [ 'aa_0', 'aa_1', 'BLOSUM62_score' ] order
 #         dict_blosum62[ aa_0, aa_1 ] = score
 
-def cressp( dir_file_protein_target = None, dir_file_protein_query = 'human', dir_folder_output = 'default', n_threads = 1, l_window_size = [ 30 ], float_thres_e_value = 1e-20, flag_use_HMM_search = False, dir_file_query_hmmdb = 'human', flag_use_rcsb_pdb_only = False, int_number_of_proteins_in_a_batch_during_dnn_prediction = 1000, flag_only_use_structural_properties_of_query_proteins = False, float_thres_avg_score_blosum_weighted__b_cell = 0.15, float_thres_avg_score_blosum__b_cell = 0.0, float_thres_rsa_correlation = 0.0, float_thres_avg_blosum62_score_for_mhc = 2, float_thres_min_mhc_allele_frequency = 0.5, float_thres_binding_affinities_in_nM = 500, flag_replace_unconventional_acid_code = False, flag_use_all_gpu_devices = False ) :
+def cressp( dir_file_protein_target = None, dir_file_protein_query = 'human', dir_folder_output = 'default', n_threads = 1, l_window_size = [ 30 ], float_thres_e_value = 30, flag_use_HMM_search = False, dir_file_query_hmmdb = 'human', flag_use_rcsb_pdb_only = False, int_number_of_proteins_in_a_batch_during_dnn_prediction = 1000, flag_only_use_structural_properties_of_query_proteins = False, float_thres_avg_score_blosum_weighted__b_cell = 0.15, float_thres_avg_score_blosum__b_cell = 0.0, float_thres_rsa_correlation = 0.0, float_thres_avg_blosum62_score_for_mhc = 2, float_thres_min_mhc_allele_frequency = 0.5, float_thres_binding_affinities_in_nM = 500, flag_replace_unconventional_acid_code = False, flag_use_all_gpu_devices = False, flag_deduplicate_based_on_aligned_subsequences_for_visualization = False ) :
     """
     The main function of Cross-Reactive-Epitope-Search-using-Structural-Properties-of-proteins (CRESSP)
     
@@ -41,7 +46,7 @@ def cressp( dir_file_protein_target = None, dir_file_protein_query = 'human', di
             
     l_window_size = [ 30 ] : list of window sizes for searching cross-reactive epitopes from the alignments between query and target proteins.
             
-    float_thres_e_value = 1e-20 : (Default: 1e-20) threshold for the global alignment e-value.
+    float_thres_e_value = 30 : (Default: 30, meaning no filtering) threshold for the global alignment e-value.
             
     flag_use_HMM_search = False : (Default: False) Set this flag to perform HMM search in addition to BLASTP search. HMM profile search is performed with HMMER3. The search usually takes several hours for metagenome-assembled genomes.
             
@@ -64,6 +69,8 @@ def cressp( dir_file_protein_target = None, dir_file_protein_query = 'human', di
     float_thres_min_mhc_allele_frequency = 0.5 : (Default: 0.5) a threshold for filtering out MHC alleles with low allele frequencies. MHC alleles with allele frequency above the threshold for at least one population will be used for cross-reactive T-cell epitope prediction
             
     float_thres_binding_affinities_in_nM = 500 : (Default: 500) a threshold predicted IC50 values for filtering predicted T-cell cross-reactive epitopes. A pair of peptides whose geometric average of predicted binding affinities (IC50) values above this threshold will be removed.
+    
+    flag_deduplicate_based_on_aligned_subsequences_for_visualization = False : (Default: False) perform the deduplication step for removing redundant predicted cross-reactive epitopes based on aligned subsequences
             
     """
     
@@ -78,6 +85,12 @@ def cressp( dir_file_protein_target = None, dir_file_protein_query = 'human', di
     """
     Parse arguments
     """
+    # define logging levels
+    logger = multiprocessing.log_to_stderr( )
+    logger.setLevel( logging.INFO )
+    Info( 'CRESSP: comparative analysis of two sets of proteomes for searching potentially cross-reactive B-cell and T-cell epitopes' )
+    Info( '[Pipeline Start] Pipeline Started at ' + TIME_GET_timestamp( True ) )
+    
     # fixed arguments
     float_search_thres_e_value = 30 # set e-value threshold for search (fixed for maximum sensitivity)
 
@@ -97,7 +110,7 @@ def cressp( dir_file_protein_target = None, dir_file_protein_query = 'human', di
         parser.add_argument( "-o", "--dir_folder_output", help = "(Default: a subdirectory 'cressp_out/' of the current directory) an output directory", default = "default" )
         parser.add_argument( "-c", "--cpu", help = "(Default: 1) Number of logical CPUs (threads) to use in the current compute node.", default = '1' )
         parser.add_argument( "-w", "--window_size", help = "(Default: 30) list of window sizes separated by comma. Example: 15,30,45", default = "30" )
-        parser.add_argument( "-e", "--float_thres_e_value", help = "(Default: 1e-20) threshold for the global alignment e-value in a scientific notation Example: 1e-3", default = "1e-20" )
+        parser.add_argument( "-e", "--float_thres_e_value", help = "(Default: 30, meaning no filtering) threshold for the global alignment e-value in a scientific notation Example: 1e-3", default = "30" )
         parser.add_argument( "-H", "--flag_use_HMM_search", help = "(Default: False) Set this flag to perform HMM search in addition to BLASTP search. HMM profile search is performed with HMMER3. The search usually takes several hours for metagenome-assembled genomes", action = 'store_true' )
         parser.add_argument( "-d", "--dir_file_query_hmmdb", help = "(Default: a HMM profile database of 1012 human proteins searched against UniProt Pan Proteomes. These proteins consist of experimentally validated human autoantigens) a file containing HMM DB of query proteins aligned against pan-proteomes", default = "human" )
         parser.add_argument( "-R", "--flag_use_rcsb_pdb_only", help = "(Default: False) When calculating consensus structural properties of input proteins, do not use homology-based modeled structures from SWISS-MODEL repositories and only use experimental protein structures from RCSB PDB", action = 'store_true' )
@@ -112,9 +125,10 @@ def cressp( dir_file_protein_target = None, dir_file_protein_query = 'human', di
         parser.add_argument( "-a", "--float_thres_binding_affinities_in_nM", help = "(Default: 500) a threshold predicted IC50 values for filtering predicted T-cell cross-reactive epitopes. A pair of peptides whose geometric average of predicted binding affinities (IC50) values above this threshold will be removed.", default = '500' )
         parser.add_argument( "-U", "--flag_replace_unconventional_acid_code", help = "(Default: False) If this flag is set, unconventional amino acids in the input protein sequences will be replaced with chemically similar amino acid. Specifically, Selenocysteine (U) to Cysteine (C), Pyrrolysine (O) to Tyrosine (Y)", action = 'store_true' )
         parser.add_argument( "-G", "--flag_use_all_gpu_devices", help = "(Default: False) Use all available GPU devices during prediction of RSA values. When this flag is set to True, the RSA prediction might be completed faster, but all GPU memories will be occupied by Tensorflow", action = 'store_true' )
+        parser.add_argument( "-D", "--flag_deduplicate_based_on_aligned_subsequences_for_visualization", help = "(Default: False) perform the deduplication step for removing redundant predicted cross-reactive epitopes based on aligned subsequences", action = 'store_true' )
         args = parser.parse_args( )
         if args.dir_file_protein_target is None :
-            print( "Required argument(s) is missing. to view help message, use -h or --help flag" )
+            Info( '[Error] Required argument(s) is missing. to view help message, use -h or --help flag' )
             sys.exit( )
 
         # parse arguments # no further processing required
@@ -135,6 +149,7 @@ def cressp( dir_file_protein_target = None, dir_file_protein_query = 'human', di
         float_thres_binding_affinities_in_nM = float( args.float_thres_binding_affinities_in_nM )
 
         # parse directory arguments
+#         dir_file_protein_target_representative
         dir_file_protein_target = args.dir_file_protein_target
         dir_file_protein_query = args.dir_file_protein_query
         dir_folder_output = args.dir_folder_output
@@ -142,7 +157,7 @@ def cressp( dir_file_protein_target = None, dir_file_protein_query = 'human', di
     else :
         ''' parse arguments when the function was called from an interactive Python interpreter '''
         if dir_file_protein_target is None :
-            print( "[CRESSP] required input 'dir_file_protein_target' was not given" )
+            Info( "[Error] required input 'dir_file_protein_target' was not given" )
             if flag_usage_from_command_line_interface : sys.exit( )
             else : return - 1
 
@@ -158,16 +173,13 @@ def cressp( dir_file_protein_target = None, dir_file_protein_query = 'human', di
             FASTA_Write( dir_file_protein_query, dict_fasta = dict_fasta_protein_query )
 
         if flag_use_HMM_search : 
-            if dir_file_query_hmmdb != 'human' :
-                print( "since query proteins is the default human proteins, the default HMM profile database will be used instead." )
-            PKG.Download_Data( "data/human/hmmdb_autoantigen.hmm.gz", dir_remote, name_package ) # download data
-            PKG.Gunzip_Data( "data/human/hmmdb_autoantigen.hmm.gz", name_package ) # unzip data
-            dir_file_query_hmmdb = pkg_resources.resource_filename( name_package, 'data/human/hmmdb_autoantigen.hmm' ) # set default 'dir_file_query_hmmdb'
+            if dir_file_query_hmmdb == 'human' :
+                PKG.Download_Data( "data/human/hmmdb_autoantigen.hmm.gz", dir_remote, name_package ) # download data
+                PKG.Gunzip_Data( "data/human/hmmdb_autoantigen.hmm.gz", name_package ) # unzip data
+                dir_file_query_hmmdb = pkg_resources.resource_filename( name_package, 'data/human/hmmdb_autoantigen.hmm' ) # set default 'dir_file_query_hmmdb'
     else :
-
-
         if flag_use_HMM_search and dir_file_query_hmmdb == 'human' :
-            print( "exiting since query proteins other than default human proteins were given, the default HMM profile database cannot be used" )
+            Info( "[Error] exiting since query proteins other than default human proteins were given, the default HMM profile database cannot be used." )
             if flag_usage_from_command_line_interface : sys.exit( )
             else : return - 1
     # get absolute paths for file arguments        
@@ -184,25 +196,15 @@ def cressp( dir_file_protein_target = None, dir_file_protein_query = 'human', di
     # handle output folder
     if dir_folder_output[ -1 ] != '/' : # last character of a directory should be '/'
         dir_folder_output += '/'
-#     if os.path.exists( dir_folder_output ) :
-#         print( "exiting since the given output directory already exists" )
-#         if flag_usage_from_command_line_interface : sys.exit( )
-#         else : return - 1
-#     else :
-    os.makedirs( dir_folder_output, exist_ok = True )
-    # create output folders for each task
+    # define folder directories for each task
     dir_folder_pipeline = f"{dir_folder_output}pipeline/"
-    os.makedirs( dir_folder_pipeline, exist_ok = True )
     dir_folder_pipeline_temp = f'{dir_folder_pipeline}temp/' 
-    os.makedirs( dir_folder_pipeline_temp, exist_ok = True )
     dir_folder_pipeline_struc = f'{dir_folder_pipeline}struc/' # create a working directory of estimating structural properties
-    os.makedirs( dir_folder_pipeline_struc, exist_ok = True )
-    
-    # define folder directories for exporting data for web applications
     dir_folder_pipeline_web = f'{dir_folder_pipeline}web_application/' # a working directory for exporting data for web applications
-    os.makedirs( dir_folder_pipeline_web, exist_ok = True )
     dir_folder_web = f'{dir_folder_output}web_application/'
-    os.makedirs( dir_folder_web, exist_ok = True )
+    # create folders
+    for dir_folder in [ dir_folder_output, dir_folder_pipeline, dir_folder_pipeline_temp, dir_folder_pipeline_struc, dir_folder_pipeline_web, dir_folder_web ] :
+        os.makedirs( dir_folder, exist_ok = True )
 
 
     """
@@ -225,7 +227,7 @@ def cressp( dir_file_protein_target = None, dir_file_protein_query = 'human', di
     name_file_protein_target = __Get_File_Name_Without_Extension__( dir_file_protein_target ) # retrieve file name containing the query proteins
     
     """ check flag """
-    dir_file_flag = f"{dir_folder_pipeline}.copying_input_files_completed.flag"
+    dir_file_flag = f"{dir_folder_pipeline}copying_input_files_completed.flag"
     if not os.path.exists( dir_file_flag ) :
         
         try :
@@ -249,6 +251,12 @@ def cressp( dir_file_protein_target = None, dir_file_protein_query = 'human', di
         """ set flag """
         with open( dir_file_flag, 'w' ) as newfile :
             newfile.write( 'completed\n' )
+    else :
+        # define directories of input protein sequences and load query protein sequences
+        dir_file_protein_query = f"{dir_folder_pipeline}protein_query.fasta" # set directory of fasta file to the new file location
+        dir_file_protein_target = f"{dir_folder_pipeline}protein_target.fasta" # set directory of fasta file to the new file location
+        dict_fasta_protein_query = FASTA_Read( dir_file_protein_query ) # load protein sequences of query (required during BLASTP output processing)
+    Info( "[Task Completion] copying input protein sequence files was completed at " + TIME_GET_timestamp( True ) )
         
     """
     Report external and internal program settings
@@ -260,6 +268,7 @@ def cressp( dir_file_protein_target = None, dir_file_protein_query = 'human', di
         'flag_use_HMM_search' : flag_use_HMM_search,
         'flag_use_rcsb_pdb_only' : flag_use_rcsb_pdb_only,
         'flag_only_use_structural_properties_of_query_proteins' : flag_only_use_structural_properties_of_query_proteins,
+        'flag_deduplicate_based_on_aligned_subsequences_for_visualization' : flag_deduplicate_based_on_aligned_subsequences_for_visualization,
         'float_thres_avg_blosum62_score_for_mhc' : float_thres_avg_blosum62_score_for_mhc,
         'float_thres_min_mhc_allele_frequency' : float_thres_min_mhc_allele_frequency,
         'l_window_size' : l_window_size,
@@ -280,14 +289,26 @@ def cressp( dir_file_protein_target = None, dir_file_protein_query = 'human', di
         'dir_folder_web' : dir_folder_web,
         'name_file_protein_query' : name_file_protein_query,
         'name_file_protein_target' : name_file_protein_target }
-    print( "cressp will be run with the following setting:" )
-    print( dict_cressp_setting )
-
+    
+    Info( f"[Setting] cressp will be run with the following setting: {str( dict_cressp_setting )}" )
+    
+    ''' export CRESSP setting '''
+    dir_file_json_setting_cressp = f"{dir_folder_pipeline}cressp_setting.json"
+    if os.path.exists( dir_file_json_setting_cressp ) :
+        with open( dir_file_json_setting_cressp, 'r' ) as file :
+            j = json.load( file )
+        if j != dict_cressp_setting :
+            Info( f"[Warning] the current CRESSP setting is different from the previous CRESSP setting recorded in the pipeline folder. The previous setting will be used." )
+            with open( dir_file_json_setting_cressp, 'r' ) as file :
+                dict_cressp_setting = json.load( file ) # override current CRESSP setting with previous CRESSP setting
+    with open( dir_file_json_setting_cressp, 'w' ) as newfile :
+        json.dump( dict_cressp_setting, newfile )    
+    
     """
     Perform BLASTP alignment
     """
     """ check flag """
-    dir_file_flag = f"{dir_folder_pipeline}.blastp_completed.flag"
+    dir_file_flag = f"{dir_folder_pipeline}blastp_completed.flag"
     if not os.path.exists( dir_file_flag ) :
 
         # create blastp_db using query_protein sequences
@@ -304,21 +325,26 @@ def cressp( dir_file_protein_target = None, dir_file_protein_query = 'human', di
         """ set flag """
         with open( dir_file_flag, 'w' ) as newfile :
             newfile.write( 'completed\n' )
-
+    else :
+        # define blastp output file
+        dir_file_blastp_output = f'{dir_folder_pipeline}blastp.tsv.gz'
+    Info( "[Task Completion] BLASTP search was completed at " + TIME_GET_timestamp( True ) )
+            
     """
     Perform HMMER alignment using a given HMM profile DB (using a couple of HMM profile DBs would be also helpful in the near future)
     """
     # run hmmsearch according to 'flag_use_HMM_search' flag
     if flag_use_HMM_search : 
         """ check flag """
-        dir_file_flag = f"{dir_folder_pipeline}.hmmsearch_completed.flag"
+        dir_file_flag = f"{dir_folder_pipeline}hmmsearch_completed.flag"
+        dir_file_hmmsearch_output = f'{dir_folder_pipeline}hmmsearch.out'
         if not os.path.exists( dir_file_flag ) :
-            dir_file_hmmsearch_output = f'{dir_folder_pipeline}hmmsearch.out'
             OS_Run( [ 'hmmsearch', '-o', dir_file_hmmsearch_output, '--acc', '--notextw', '--cpu', f'{n_threads}', '-E', f'{float_search_thres_e_value}', dir_file_query_hmmdb, dir_file_protein_target ], dir_file_stdout = f"{dir_folder_pipeline}hmmsearch.stdout.txt", dir_file_stderr = f"{dir_folder_pipeline}hmmsearch.stderr.txt", return_output = False ) # run hmmsearch
             """ set flag """
             with open( dir_file_flag, 'w' ) as newfile :
                 newfile.write( 'completed\n' )
-
+        Info( "[Task Completion] HMMER search was completed at " + TIME_GET_timestamp( True ) )
+    
     """
     Combine BLASTP and HMMSEARCH outputs 
     """
@@ -330,6 +356,7 @@ def cressp( dir_file_protein_target = None, dir_file_protein_query = 'human', di
         # load blastp result
         dict_qacc_to_seq = FASTA_Read( dir_file_protein_target ) # read query protein sequences
         dict_qacc_to_seq = dict( ( header.split( ' ', 1 )[ 0 ], dict_qacc_to_seq[ header ] ) for header in list( dict_qacc_to_seq ) )
+        set_acc_target = set( dict_qacc_to_seq ) # retrieve a set of acc_target
         df_blastp = BLAST_Read( dir_file_blastp_output, dict_qaccver_to_seq = dict_qacc_to_seq ) 
         df_blastp = df_blastp[ [ 'saccver', 'qaccver', 'sstart', 'send', 'qstart', 'qend', 'subject_seq_aligned', 'query_seq_aligned', 'evalue', 'pident' ] ]
         df_blastp.pident = df_blastp.pident / 100
@@ -338,8 +365,9 @@ def cressp( dir_file_protein_target = None, dir_file_protein_query = 'human', di
 
         # load hmmer result according to 'flag_use_HMM_search' flag
         if flag_use_HMM_search : 
-            df = HMMER_HMMSEARCH_Read_output( dir_file_hmmsearch_output )
             dict_qacc_to_seq = dict( ( header.split( ' ', 1 )[ 0 ], dict_fasta_protein_query[ header ] ) for header in list( dict_fasta_protein_query ) )
+            # ignore if records from profiles that do not match input query protein sequences 
+            df = PD_Select( HMMER_HMMSEARCH_Read_output( dir_file_hmmsearch_output ), query_accession = set( dict_qacc_to_seq ) ) 
             l_query_alignment = list( ) # replace query consensus sequence with query sequence
             for query_accession, query_alignment, query_start, query_end in df[ [ 'query_accession', 'query_alignment', 'query_start', 'query_end' ] ].values :
                 query_seq = dict_qacc_to_seq[ query_accession ][ query_start - 1 : query_end ] # retrive a subsequence of query sequence
@@ -356,6 +384,8 @@ def cressp( dir_file_protein_target = None, dir_file_protein_query = 'human', di
             df_hmmer = df_hmmer[ [ 'query_accession', 'target_accession', 'query_start', 'query_end', 'target_start', 'target_end', 'query_alignment', 'target_alignment', 'conditional_Evalue', 'accuracy' ] ] # subset common columns
             df_hmmer.columns = [ 'query_accession', 'target_accession', 'query_start', 'query_end', 'target_start', 'target_end', 'query_alignment', 'target_alignment', 'e_value', 'identity' ] # rename columns
             df_hmmer[ 'source' ] = 'hmmer' 
+            set_acc_query = set( header.split( ' ', 1 )[ 0 ] for header in dict_fasta_protein_query ) # retrieve a set of acc_query
+            df_hmmer = PD_Select( df_hmmer, query_accession = set_acc_query ) # exclude records with accessions that does not exist in the input query protein sequences
 
         df_matched = pd.concat( [ df_hmmer, df_blastp ], ignore_index = True ) if flag_use_HMM_search else df_blastp
         df_matched.to_csv( dir_file_matched, sep = '\t', index = False )
@@ -363,13 +393,14 @@ def cressp( dir_file_protein_target = None, dir_file_protein_query = 'human', di
         """ set flag """
         with open( dir_file_flag, 'w' ) as file :
             file.write( f"search results were written at {TIME_GET_timestamp( )}" )
-
+    Info( "[Task Completion] Combining HMMER and BLASTP search results was completed at " + TIME_GET_timestamp( True ) )
+    
     """
     Estimate structural properties of proteins 
     """
     
     """ check flag """
-    dir_file_flag = f"{dir_file_matched}.rsa_estimation_complated.flag"
+    dir_file_flag = f"{dir_folder_pipeline}rsa_estimation_completed.flag"
     if not os.path.exists( dir_file_flag ) :
         # use previously calculated structural properties when the default query proteins were used
         if flag_default_protein_query_was_used :
@@ -392,7 +423,7 @@ def cressp( dir_file_protein_target = None, dir_file_protein_query = 'human', di
         """ set flag """
         with open( dir_file_flag, 'w' ) as newfile :
             newfile.write( 'completed\n' )
-
+    Info( "[Task Completion] Estimation of structural properties of input protein sequences completed at " + TIME_GET_timestamp( True ) )
 
     """
     Calculate B-cell epitope similarity scores based on structural properties of proteins 
@@ -407,6 +438,7 @@ def cressp( dir_file_protein_target = None, dir_file_protein_query = 'human', di
         """ set flag """
         with open( dir_file_flag, 'w' ) as newfile :
             newfile.write( 'completed\n' )
+    Info( "[Task Completion] Prediction of B-cell cross-reactivity was completed at " + TIME_GET_timestamp( True ) )
     
     """ 
     Calculate T-cell epitope similarity scores based on BLOSUM62 scores and predicted binding affinity scores.
@@ -421,6 +453,7 @@ def cressp( dir_file_protein_target = None, dir_file_protein_query = 'human', di
         """ set flag """
         with open( dir_file_flag, 'w' ) as newfile :
             newfile.write( 'completed\n' )
+    Info( "[Task Completion] Prediction of T-cell cross-reactivity was completed at " + TIME_GET_timestamp( True ) )
     
     """
     Further process and export data for visualization using a web application
@@ -431,9 +464,7 @@ def cressp( dir_file_protein_target = None, dir_file_protein_query = 'human', di
     if not os.path.exists( dir_file_flag ) :
 
         # combine results of all 'window_size' values
-        df = pd.concat( list( pd.read_csv( dir_file, sep = '\t', low_memory = False ) for dir_file in glob.glob( f"{dir_folder_pipeline}b_cell.subsequence__window_size_*.tsv.gz" ) ), ignore_index = True )
-        df.to_csv( f"{dir_folder_pipeline}b_cell.subsequence.tsv.gz", sep = '\t', index = False )
-        del df
+        OS_FILE_Combine_Files_in_order( glob.glob( f"{dir_folder_pipeline}b_cell.subsequence__window_size_*.tsv.gz" ), f"{dir_folder_pipeline}b_cell.subsequence.tsv.gz", remove_n_lines = 1, flag_use_header_from_first_file = True )
         # prepare data for web application using the combined subsequence
         # copy data for web application and encode using base64 encoding, and write metadata
         Prepare_data_for_web_application( f"{dir_folder_pipeline}b_cell.subsequence.tsv.gz", f"{dir_folder_pipeline}t_cell.mhc_binding.tsv.gz", dict_cressp_setting )
@@ -441,35 +472,50 @@ def cressp( dir_file_protein_target = None, dir_file_protein_query = 'human', di
         """ set flag """
         with open( dir_file_flag, 'w' ) as newfile :
             newfile.write( 'completed\n' )
-    
+    Info( "[Task Completion] Exporting data for CRESSP web application was completed at " + TIME_GET_timestamp( True ) )
 
 
-def Parse_Structural_Properties( dir_file_df_sp, name_col_for_identifying_protein = 'id_protein' ) :
+def Parse_Structural_Properties( dir_file_df_sp, int_datatype = None, name_col_for_identifying_protein = 'id_protein' ) :
     """
     # 2021-05-31 20:29:02 
     parse structural properties with typical parameters for parsing ascii-encoded structural properties and typical column names  
     
     'dir_file_df_sp' : a file directory to TSV file containing structural properties estimated from CRESSP. Alternatively, a dataframe of the TSV file can be given as an input.
-    
+    'int_datatype' : used for initializing 'datatype_acc' if 'rsa_datatype___ascii_encoding_1_character_from_33_to_36__states_Pred_Model_PDB' column does not exist
     'name_col_for_identifying_protein' : columne name of the given tabular data for identifying a unique protein. The unique value in the column will be used as a 'key' in the returned dicitonaries
     
     """
     df_sp = pd.read_csv( dir_file_df_sp, sep = '\t' ) if isinstance( dir_file_df_sp, ( str ) ) else dir_file_df_sp # read database of structural properties # if an object that is not a string datatype is given, assumes 'dir_file_df_sp' is a dataframe containing structural properties
+    df_sp.set_index( name_col_for_identifying_protein, inplace = True )
     
     dict_kw_rsa = dict( ascii_min = 33, ascii_max = 126, l_ascii_to_exclude = [ 62 ], n_char = 2, value_min = 0, value_max = 1 )
     dict_kw_torsion_angle = dict( ascii_min = 33, ascii_max = 126, l_ascii_to_exclude = [ 62 ], n_char = 2, value_min = -180, value_max = 180 )
     dict_kw_ss8 = dict( ascii_min = 33, ascii_max = 41, l_ascii_to_exclude = [ 62 ], n_char = 1, value_min = 0, value_max = 8 )
     dict_kw_datatype = dict( ascii_min = 33, ascii_max = 36, l_ascii_to_exclude = [ 62 ], n_char = 1, value_min = 0, value_max = 3 )
 
+    
     dict_sp = dict( )
-    dict_sp[ 'acc' ] = ASCII_Decode( df_sp.set_index( name_col_for_identifying_protein ).rsa___ascii_encoding_2_characters_from_33_to_126__from_0_to_1.to_dict( ), ** dict_kw_rsa ) # decode mkdssp outputs of RCSB PDB data
-    dict_sp[ 'phi' ] = ASCII_Decode( df_sp.set_index( name_col_for_identifying_protein )[ 'phi___ascii_encoding_2_characters_from_33_to_126__from_-180_to_180' ].to_dict( ), ** dict_kw_torsion_angle )
-    dict_sp[ 'psi' ] = ASCII_Decode( df_sp.set_index( name_col_for_identifying_protein )[ 'psi___ascii_encoding_2_characters_from_33_to_126__from_-180_to_180' ].to_dict( ), ** dict_kw_torsion_angle )
-    dict_sp[ 'ss8' ] = ASCII_Decode( df_sp.set_index( name_col_for_identifying_protein )[ 'ss8___ascii_encoding_1_character_from_33_to_41__states_G_H_I_E_B_T_S_C' ].to_dict( ), ** dict_kw_ss8 )
-    dict_sp[ 'datatype_acc' ] = ASCII_Decode( df_sp.set_index( name_col_for_identifying_protein )[ 'rsa_datatype___ascii_encoding_1_character_from_33_to_36__states_Pred_Model_PDB' ].to_dict( ), ** dict_kw_datatype )
-    dict_fasta = df_sp.set_index( name_col_for_identifying_protein )[ 'structure_id___redundancy_reduced' ].dropna( ).to_dict( )
-    dict_sp[ 'structure_id' ] = dict( ( acc, Decode_List_of_Strings( dict_fasta[ acc ] ) ) for acc in dict_fasta )
-    dict_sp[ 'seq' ] = df_sp.set_index( name_col_for_identifying_protein )[ 'seq' ].to_dict( )
+    dict_sp[ 'acc' ] = ASCII_Decode( df_sp.rsa___ascii_encoding_2_characters_from_33_to_126__from_0_to_1.to_dict( ), ** dict_kw_rsa ) # decode mkdssp outputs of RCSB PDB data
+    dict_sp[ 'phi' ] = ASCII_Decode( df_sp[ 'phi___ascii_encoding_2_characters_from_33_to_126__from_-180_to_180' ].to_dict( ), ** dict_kw_torsion_angle )
+    dict_sp[ 'psi' ] = ASCII_Decode( df_sp[ 'psi___ascii_encoding_2_characters_from_33_to_126__from_-180_to_180' ].to_dict( ), ** dict_kw_torsion_angle )
+    dict_sp[ 'ss8' ] = ASCII_Decode( df_sp[ 'ss8___ascii_encoding_1_character_from_33_to_41__states_G_H_I_E_B_T_S_C' ].to_dict( ), ** dict_kw_ss8 )
+    if 'rsa_datatype___ascii_encoding_1_character_from_33_to_36__states_Pred_Model_PDB' in df_sp.columns.values :
+        dict_sp[ 'datatype_acc' ] = ASCII_Decode( df_sp[ 'rsa_datatype___ascii_encoding_1_character_from_33_to_36__states_Pred_Model_PDB' ].to_dict( ), ** dict_kw_datatype )
+    else :
+        if int_datatype is not None :
+            return -1
+            
+        # initialization of a dictionary of arrays containing the identifier of the current dataset of 'acc' datatype
+        dict_datatype_acc = dict( )
+        for h in dict_sp[ 'acc' ] :
+            arr = np.zeros_like( dict_sp[ 'acc' ][ h ] ) # default = 0 (predicted structural property)
+            arr[ ~ np.isnan( dict_sp[ 'acc' ][ h ] ) ] = int_datatype
+            dict_datatype_acc[ h ] = arr
+        dict_sp[ 'datatype_acc' ] = dict_datatype_acc
+    if 'structure_id___redundancy_reduced' in df_sp.columns.values :
+        dict_fasta = df_sp[ 'structure_id___redundancy_reduced' ].dropna( ).to_dict( )
+        dict_sp[ 'structure_id' ] = dict( ( acc, Decode_List_of_Strings( dict_fasta[ acc ] ) ) for acc in dict_fasta )
+        dict_sp[ 'seq' ] = df_sp[ 'seq' ].to_dict( )
     
     return dict_sp
 
