@@ -127,7 +127,7 @@ def Predict_T_cell_cross_reactivity( dir_folder_pipeline, float_thres_avg_blosum
             df_mhc_web = pd.DataFrame( l, columns = [ 'index', 'mhc_allele', 'affinity_query', 'affinity_target' ] ).set_index( 'index' ).join( df_subsequence_mhc_i, how = 'left' ) 
             df_mhc_web[ 'mhc_class' ] = 1 # MHC-Class is I
             df_mhc_web[ 'score_for_sorting' ] = df_mhc_web.score_blosum / ( df_mhc_web.affinity_query * df_mhc_web.affinity_target ) ** 0.5 # calculate score for sorting
-            df_mhc_web.to_csv( newfile, sep = '\t', index = False, header = False )
+            df_mhc_web.to_csv( newfile, sep = '\t', index = False, header = False, mode = 'wb' )
 
             ''' complete the operation if the end of the file has been reached '''
             if flag_file_end_reached :
